@@ -82,6 +82,13 @@ enum {
 	PROP_REMOTES
 };
 
+enum {
+	CHANGED,
+	LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = { 0, };
+
 #define GET_PRIV(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GIGGLE_TYPE_GIT, GiggleGitPriv))
 
 static void
@@ -135,6 +142,15 @@ giggle_git_class_init (GiggleGitClass *class)
 							      "The remote sources",
 							      NULL, G_PARAM_READABLE));
 
+	signals[CHANGED] =
+		g_signal_new ("changed",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (GiggleGitClass, changed),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__VOID,
+			      G_TYPE_NONE, 0);
+
 	g_type_class_add_private (object_class, sizeof (GiggleGitPriv));
 }
 
@@ -370,14 +386,14 @@ static gchar *
 giggle_git_get_description_file (const GiggleGit *git)
 {
 	GiggleGitPriv *priv = GET_PRIV (git);
-	
+
 	return g_build_filename (priv->git_dir, "description", NULL);
 }
 
 static void
 giggle_git_update_description (GiggleGit *git)
 {
-	// FIXME: read .git/description into description; install a file watch
+	/* FIXME: read .git/description into description; install a file watch */
 	GiggleGitPriv *priv;
 	GError        *error;
 	gchar* description;
@@ -455,7 +471,7 @@ giggle_git_get_description (GiggleGit *git)
 
 void
 giggle_git_write_description (GiggleGit    *git,
-			      const gchar  *description) // FIXME: add GError?
+			      const gchar  *description) /* FIXME: add GError? */
 {
 	GiggleGitPriv *priv;
 	GError        *error;
@@ -535,7 +551,8 @@ giggle_git_set_directory (GiggleGit    *git,
 		priv->project_dir = NULL;
 	} else {
 		/* .../giggle/.git
-		 *            ^    */
+		 *            ^
+		 */
 		if(*(--suffix) == G_DIR_SEPARATOR) {
 			*suffix = '\0';
 			priv->project_dir = g_strdup (tmp_dir);
@@ -555,7 +572,7 @@ giggle_git_set_directory (GiggleGit    *git,
 		if (suffix) {
 			*suffix = '\0';
 			tmp_dir = g_path_get_basename (priv->git_dir);
-			*suffix = '.'; // restore
+			*suffix = '.'; /* restore */
 		} else {
 			tmp_dir = NULL;
 		}
@@ -652,10 +669,10 @@ giggle_git_run_job_full (GiggleGit             *git,
 
 	if (giggle_job_get_command_line (job, &command)) {
 		GitJobData    *data;
-		
+
 		data = g_slice_new0 (GitJobData);
 		data->id = giggle_dispatcher_execute (priv->dispatcher,
-						      priv->directory,
+						      priv->project_dir,
 						      command,
 						      (GiggleExecuteCallback) git_execute_callback,
 						      git);
@@ -703,3 +720,10 @@ giggle_git_cancel_job (GiggleGit *git, GiggleJob *job)
 	g_hash_table_remove (priv->jobs, GINT_TO_POINTER (id));
 }
 
+void
+giggle_git_changed (GiggleGit *git)
+{
+	g_return_if_fail (GIGGLE_IS_GIT (git));
+
+	g_signal_emit (git, signals[CHANGED], 0);
+}
