@@ -27,6 +27,7 @@
 #include "giggle-git.h"
 #include "giggle-git-revisions.h"
 #include "giggle-git-refs.h"
+#include "giggle-git-diff.h"
 #include "giggle-view-history.h"
 #include "giggle-file-list.h"
 #include "giggle-revision-list.h"
@@ -52,16 +53,22 @@ struct GiggleViewHistoryPriv {
 	GtkWidget *revision_hpaned;
 
 	GtkWidget *revision_view_expander;
+	GtkWidget *revision_expander_label;
+	GtkWidget *branches_label;
+
 	GtkWidget *diff_view_expander;
 	GtkWidget *diff_view_sw;
 
 	GiggleGit *git;
 	GiggleJob *job;
+	GiggleJob *diff_current_job;
 
 	GList     *history; /* reversed list of history elems */
 	GList     *current_history_elem;
 
-	guint     compact_mode : 1;
+	guint      selection_changed_idle;
+
+	guint      compact_mode : 1;
 };
 
 static void     view_history_finalize              (GObject *object);
@@ -76,6 +83,8 @@ static void     view_history_revision_list_selection_changed_cb (GiggleRevisionL
 static gboolean view_history_revision_list_key_press_cb         (GiggleRevisionList *list,
 								 GdkEventKey        *event,
 								 GiggleViewHistory  *view);
+static void     view_history_file_list_status_changed           (GiggleFileList     *list,
+								 GiggleViewHistory  *view);
 
 static gboolean view_history_search                             (GiggleSearchable      *searchable,
 								 const gchar           *search_term,
@@ -135,6 +144,109 @@ giggle_view_history_history_init (GiggleHistoryIface *iface)
 	iface->can_go_forward = view_history_can_go_forward;
 }
 
+static void
+view_history_expander_label_size_allocation (GtkWidget         *widget,
+					     GtkAllocation     *allocation,
+					     GiggleViewHistory *view)
+{
+	GiggleViewHistoryPriv *priv;
+	GtkAllocation          expander_allocation;
+
+	priv = GET_PRIV (view);
+
+	/* terrible, but there's no other way to show
+	 * something in the label area aligned to the right
+	 */
+	expander_allocation = priv->revision_view_expander->allocation;
+	allocation->width = expander_allocation.width -
+		(allocation->x - expander_allocation.x);
+
+	gtk_widget_size_allocate (widget, allocation);
+}
+
+static GtkWidget *
+view_history_create_revision_expander (GiggleViewHistory *view)
+{
+	GiggleViewHistoryPriv *priv;
+	GtkWidget             *expander, *label;
+
+	priv = GET_PRIV (view);
+
+	priv->revision_expander_label = gtk_hbox_new (FALSE, 12);
+
+	label = gtk_label_new_with_mnemonic (_("Revision _information"));
+	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
+	gtk_box_pack_start (GTK_BOX (priv->revision_expander_label), label, FALSE, FALSE, 0);
+
+	priv->branches_label = gtk_label_new (NULL);
+	gtk_misc_set_alignment (GTK_MISC (priv->branches_label), 1.0, 0.0);
+	gtk_label_set_ellipsize (GTK_LABEL (priv->branches_label), PANGO_ELLIPSIZE_END);
+	gtk_box_pack_start (GTK_BOX (priv->revision_expander_label), priv->branches_label, TRUE, TRUE, 0);
+
+	expander = gtk_expander_new (NULL);
+	gtk_expander_set_label_widget (GTK_EXPANDER (expander), priv->revision_expander_label);
+	gtk_widget_show_all (expander);
+
+	g_signal_connect (priv->revision_expander_label, "size-allocate",
+			  G_CALLBACK (view_history_expander_label_size_allocation), view);
+
+	return expander;
+}
+
+static void
+view_history_git_changed (GiggleViewHistory *view)
+{
+	GiggleViewHistoryPriv *priv;
+
+	priv = GET_PRIV (view);
+
+	/* clear history */
+	g_list_foreach (priv->history, (GFunc) g_free, NULL);
+	g_list_free (priv->history);
+	priv->history = NULL;
+	priv->current_history_elem = NULL;
+
+	view_history_update_revisions (view);
+}
+
+static void
+view_history_set_busy (GtkWidget *widget,
+		       gboolean   busy)
+{
+	if (!GTK_WIDGET_REALIZED (widget)) {
+		return;
+	}
+
+	if (busy) {
+		GdkCursor *cursor;
+
+		cursor = gdk_cursor_new (GDK_WATCH);
+		gdk_window_set_cursor (widget->window, cursor);
+		gdk_cursor_unref (cursor);
+	} else {
+		gdk_window_set_cursor (widget->window, NULL);
+	}
+}
+
+static void
+view_history_git_dir_notify (GiggleViewHistory *view)
+{
+	GiggleViewHistoryPriv *priv;
+	GtkTextBuffer         *buffer;
+
+	priv = GET_PRIV (view);
+
+	view_history_update_revisions (view);
+	view_history_set_busy (priv->file_list, TRUE);
+
+	/* empty views */
+	giggle_diff_tree_view_set_revisions (GIGGLE_DIFF_TREE_VIEW (priv->diff_tree_view), NULL, NULL);
+	giggle_revision_view_set_revision (GIGGLE_REVISION_VIEW (priv->revision_view), NULL);
+
+	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->diff_view));
+	gtk_text_buffer_set_text (buffer, "", -1);
+}
+
 static void
 giggle_view_history_init (GiggleViewHistory *view)
 {
@@ -162,7 +274,7 @@ giggle_view_history_init (GiggleViewHistory *view)
 	priv->revision_hpaned = gtk_hpaned_new ();
 	gtk_widget_show (priv->revision_hpaned);
 	gtk_paned_pack2 (GTK_PANED (priv->main_vpaned), priv->revision_hpaned, FALSE, FALSE);
-	
+
 	/* diff file view */
 	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
 	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
@@ -196,6 +308,8 @@ giggle_view_history_init (GiggleViewHistory *view)
 
 	g_signal_connect (priv->file_list, "path-selected",
 			  G_CALLBACK (view_history_path_selected), view);
+	g_signal_connect (priv->file_list, "status-changed",
+			  G_CALLBACK (view_history_file_list_status_changed), view);
 
 	gtk_paned_pack1 (GTK_PANED (priv->main_hpaned), priv->file_list_sw, FALSE, FALSE);
 
@@ -217,7 +331,7 @@ giggle_view_history_init (GiggleViewHistory *view)
 	gtk_paned_pack1 (GTK_PANED (priv->main_vpaned), scrolled_window, TRUE, FALSE);
 
 	/* revision view */
-	priv->revision_view_expander = gtk_expander_new_with_mnemonic (_("Revision _information"));
+	priv->revision_view_expander = view_history_create_revision_expander (view);
 	priv->revision_view = giggle_revision_view_new ();
 	gtk_container_add (GTK_CONTAINER (priv->revision_view_expander), priv->revision_view);
 	gtk_expander_set_expanded (GTK_EXPANDER (priv->revision_view_expander), TRUE);
@@ -247,8 +361,9 @@ giggle_view_history_init (GiggleViewHistory *view)
 	/* git interaction */
 	priv->git = giggle_git_get ();
 	g_signal_connect_swapped (priv->git, "notify::git-dir",
-				  G_CALLBACK (view_history_update_revisions), view);
-	view_history_update_revisions (view);
+				  G_CALLBACK (view_history_git_dir_notify), view);
+	g_signal_connect_swapped (priv->git, "changed",
+				  G_CALLBACK (view_history_git_changed), view);
 }
 
 static void
@@ -264,6 +379,12 @@ view_history_finalize (GObject *object)
 		priv->job = NULL;
 	}
 
+	if (priv->diff_current_job) {
+		giggle_git_cancel_job (priv->git, priv->diff_current_job);
+		g_object_unref (priv->diff_current_job);
+		priv->diff_current_job = NULL;
+	}
+
 	g_list_foreach (priv->history, (GFunc) g_free, NULL);
 	g_list_free (priv->history);
 
@@ -272,33 +393,106 @@ view_history_finalize (GObject *object)
 	G_OBJECT_CLASS (giggle_view_history_parent_class)->finalize (object);
 }
 
+typedef struct {
+	GiggleViewHistory *view;
+	GiggleRevision    *revision1;
+	GiggleRevision    *revision2;
+} ViewHistorySelectionIdleData;
+
+static gboolean
+view_history_selection_changed_idle (ViewHistorySelectionIdleData *data)
+{
+	GiggleViewHistoryPriv *priv;
+	GList                 *files;
+
+	GDK_THREADS_ENTER ();
+
+	priv = GET_PRIV (data->view);
+	files = NULL;
+
+	if (priv->current_history_elem) {
+		files = g_list_prepend (NULL, g_strdup ((gchar *) priv->current_history_elem->data));
+	}
+
+	giggle_diff_view_set_revisions (GIGGLE_DIFF_VIEW (priv->diff_view),
+					data->revision1, data->revision2, files);
+
+	giggle_diff_tree_view_set_revisions (GIGGLE_DIFF_TREE_VIEW (priv->diff_tree_view),
+					     data->revision1, data->revision2);
+	giggle_file_list_highlight_revisions (GIGGLE_FILE_LIST (priv->file_list),
+					      data->revision1, data->revision2);
+
+	GDK_THREADS_LEAVE ();
+
+	return FALSE;
+}
+
+static void
+view_history_set_branches_label (GiggleViewHistory *view,
+				 GiggleRevision    *revision)
+{
+	GiggleViewHistoryPriv *priv;
+	GiggleRef             *ref;
+	GList                 *branches;
+	GString               *str;
+
+	priv = GET_PRIV (view);
+
+	gtk_label_set_text (GTK_LABEL (priv->branches_label), NULL);
+
+	if (!revision) {
+		return;
+	}
+
+	branches = giggle_revision_get_descendent_branches (revision);
+
+	if (branches) {
+		str = g_string_new ("");
+
+		g_string_append_printf (str, "<b>%s</b>:",
+					ngettext ("Branch", "Branches",
+						  g_list_length (branches)));
+
+		while (branches) {
+			ref = GIGGLE_REF (branches->data);
+			g_string_append_printf (str, " %s", giggle_ref_get_name (ref));
+			branches = branches->next;
+		}
+
+		gtk_label_set_markup (GTK_LABEL (priv->branches_label), str->str);
+		g_string_free (str, TRUE);
+	}
+}
+
 static void
 view_history_revision_list_selection_changed_cb (GiggleRevisionList *list,
 						 GiggleRevision     *revision1,
 						 GiggleRevision     *revision2,
 						 GiggleViewHistory  *view)
 {
-	GiggleViewHistoryPriv *priv;
-	GList                 *files;
+	GiggleViewHistoryPriv        *priv;
+	ViewHistorySelectionIdleData *data;
 
 	priv = GET_PRIV (view);
-	files = NULL;
+
+	view_history_set_branches_label (view, revision1);
 
 	giggle_revision_view_set_revision (
 		GIGGLE_REVISION_VIEW (priv->revision_view), revision1);
 
-	if (revision1 && revision2) {
-		if (priv->current_history_elem) {
-			files = g_list_prepend (NULL, g_strdup ((gchar *) priv->current_history_elem->data));
-		}
-
-		giggle_diff_view_set_revisions (GIGGLE_DIFF_VIEW (priv->diff_view),
-						revision1, revision2, files);
-		giggle_diff_tree_view_set_revisions (GIGGLE_DIFF_TREE_VIEW (priv->diff_tree_view),
-						     revision1, revision2);
-		giggle_file_list_highlight_revisions (GIGGLE_FILE_LIST (priv->file_list),
-						revision1, revision2);
+	if (priv->selection_changed_idle) {
+		g_source_remove (priv->selection_changed_idle);
 	}
+
+	data = g_new0 (ViewHistorySelectionIdleData, 1);
+	data->view = view;
+	data->revision1 = revision1;
+	data->revision2 = revision2;
+
+	priv->selection_changed_idle =
+		g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
+				 (GSourceFunc) view_history_selection_changed_idle,
+				 data, g_free);
 }
 
 static gboolean
@@ -315,7 +509,7 @@ view_history_revision_list_key_press_cb (GiggleRevisionList *list,
 	if (event->keyval == GDK_space ||
 	    event->keyval == GDK_BackSpace) {
 		gtk_expander_set_expanded (GTK_EXPANDER (priv->diff_view_expander), TRUE);
-		
+
 		adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (priv->diff_view_sw));
 
 		value = (event->keyval == GDK_space) ?
@@ -332,6 +526,13 @@ view_history_revision_list_key_press_cb (GiggleRevisionList *list,
 	return FALSE;
 }
 
+static void
+view_history_file_list_status_changed (GiggleFileList    *list,
+				       GiggleViewHistory *view)
+{
+	view_history_set_busy (GTK_WIDGET (list), FALSE);
+}
+
 static gboolean
 view_history_search (GiggleSearchable      *searchable,
 		     const gchar           *search_term,
@@ -378,26 +579,24 @@ view_history_add_refs (GiggleRevision *revision,
 		       GList          *list,
 		       AddRefFunc      func)
 {
-	GiggleRef *ref;
-	gchar     *sha1, *sha2;
-	gboolean   updated = FALSE;
+	GiggleRef   *ref;
+	const gchar *sha1, *sha2;
+	gboolean     updated = FALSE;
 
-	g_object_get (revision, "sha", &sha1, NULL);
+	sha1 = giggle_revision_get_sha (revision);
 
 	while (list) {
 		ref = GIGGLE_REF (list->data);
-		g_object_get (ref, "sha", &sha2, NULL);
+		sha2 = giggle_ref_get_sha (ref);
 
 		if (strcmp (sha1, sha2) == 0) {
 			updated = TRUE;
 			(* func) (revision, ref);
 		}
 
-		g_free (sha2);
 		list = list->next;
 	}
 
-	g_free (sha1);
 	return updated;
 }
 
@@ -443,16 +642,19 @@ view_history_get_branches_cb (GiggleGit    *git,
 					    REVISION_COL_OBJECT, &revision,
 					    -1);
 
-			changed = view_history_add_refs (revision, branches, giggle_revision_add_branch_head);
-			changed |= view_history_add_refs (revision, tags, giggle_revision_add_tag);
+			if (revision) {
+				changed = view_history_add_refs (revision, branches, giggle_revision_add_branch_head);
+				changed |= view_history_add_refs (revision, tags, giggle_revision_add_tag);
 
-			if (changed) {
-				path = gtk_tree_model_get_path (model, &iter);
-				gtk_tree_model_row_changed (model, path, &iter);
-				gtk_tree_path_free (path);
+				if (changed) {
+					path = gtk_tree_model_get_path (model, &iter);
+					gtk_tree_model_row_changed (model, path, &iter);
+					gtk_tree_path_free (path);
+				}
+
+				g_object_unref (revision);
 			}
 
-			g_object_unref (revision);
 			valid = gtk_tree_model_iter_next (model, &iter);
 		}
 	}
@@ -461,6 +663,40 @@ view_history_get_branches_cb (GiggleGit    *git,
 	priv->job = NULL;
 }
 
+static void
+view_history_diff_current_cb (GiggleGit *git,
+			      GiggleJob *job,
+			      GError    *error,
+			      gpointer   user_data)
+{
+	GiggleViewHistoryPriv *priv;
+	GtkTreeModel          *model;
+	GtkTreePath           *path;
+	GtkTreeIter            iter;
+	const gchar           *text;
+
+	priv = GET_PRIV (user_data);
+
+	/* FIXME: error report missing */
+	if (error) {
+		return;
+	}
+
+	text = giggle_git_diff_get_result (GIGGLE_GIT_DIFF (job));
+
+	if (text && *text) {
+		model = gtk_tree_view_get_model (GTK_TREE_VIEW (priv->revision_list));
+		gtk_list_store_insert (GTK_LIST_STORE (model), &iter, 0);
+		gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+				    REVISION_COL_OBJECT, NULL,
+				    -1);
+
+		path = gtk_tree_model_get_path (model, &iter);
+		gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (priv->revision_list),
+					      path, NULL, FALSE, 0.0, 0.0);
+	}
+}
+
 static void
 view_history_get_revisions_cb (GiggleGit    *git,
 			       GiggleJob    *job,
@@ -502,6 +738,7 @@ view_history_get_revisions_cb (GiggleGit    *git,
 			revisions = revisions->next;
 		}
 
+		view_history_set_busy (GTK_WIDGET (priv->revision_list), FALSE);
 		giggle_revision_list_set_model (GIGGLE_REVISION_LIST (priv->revision_list),
 						GTK_TREE_MODEL (store));
 		g_object_unref (store);
@@ -516,6 +753,20 @@ view_history_get_revisions_cb (GiggleGit    *git,
 				    view_history_get_branches_cb,
 				    view);
 
+		/* and current diff row */
+		if (priv->diff_current_job) {
+			giggle_git_cancel_job (priv->git, priv->diff_current_job);
+			g_object_unref (priv->diff_current_job);
+			priv->diff_current_job = NULL;
+		}
+
+		priv->diff_current_job = giggle_git_diff_new ();
+
+		giggle_git_run_job (priv->git,
+				    priv->diff_current_job,
+				    view_history_diff_current_cb,
+				    view);
+
 		giggle_history_changed (GIGGLE_HISTORY (view));
 	}
 }
@@ -528,8 +779,10 @@ view_history_update_revisions (GiggleViewHistory  *view)
 
 	priv = GET_PRIV (view);
 
+	view_history_set_busy (GTK_WIDGET (priv->revision_list), TRUE);
 	giggle_revision_list_set_model (GIGGLE_REVISION_LIST (priv->revision_list), NULL);
 
+	/* get revision list */
 	if (priv->job) {
 		giggle_git_cancel_job (priv->git, priv->job);
 		g_object_unref (priv->job);
@@ -641,6 +894,20 @@ giggle_view_history_get_file_list_visible (GiggleViewHistory *view)
 	return GTK_WIDGET_VISIBLE (priv->file_list_sw);
 }
 
+void
+giggle_view_history_set_graph_visible (GiggleViewHistory *view,
+				       gboolean           visible)
+{
+	GiggleViewHistoryPriv *priv;
+
+	g_return_if_fail (GIGGLE_IS_VIEW_HISTORY (view));
+
+	priv = GET_PRIV (view);
+
+	giggle_revision_list_set_graph_visible (
+		GIGGLE_REVISION_LIST (priv->revision_list), visible);
+}
+
 static void
 view_history_go_back (GiggleHistory *history)
 {
